#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import tensorflow as tf

from merlin.models.tf.blocks.sampling.in_batch import InBatchSampler


def test_in_batch_sampler_reload():
    batch_size = 43
    sampler = InBatchSampler(batch_size=batch_size)
    serialized = tf.keras.layers.serialize(sampler)
    reloaded = tf.keras.layers.deserialize(serialized)
    assert reloaded.batch_size == batch_size
