#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import merlin.models.torch as ml


def test_mlp_block(tabular_schema, torch_tabular_data):
    tab_module = ml.TabularFeatures.from_schema(
        tabular_schema, max_sequence_length=20, aggregation="concat"
    )

    block = tab_module >> ml.MLPBlock([64, 32])

    outputs = block(torch_tabular_data)

    assert outputs.ndim == 2
    assert outputs.shape[-1] == 32
